/* file "parse.y" */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file. */

/* specification parser */
%{
#include <stdlib.h>
#include <suif_copyright.h>
#include <string.h>
#include <ctype.h>
#include <suif1/misc.h>
#include <suif1/glist.h>
#include "symbol.h"
#include "sstring.h"
#include "lex.h"
extern int yylex();
extern void yyerror(const char *);
extern Symbol error_sym;
Symbol_table *symtab;
i_stream *current_input;
char *cmdnam;
int n_errors, n_warnings;
extern Symbol *problem_name;
extern SV_dflow *problem;
extern int yyparse();
extern void set_name(Symbol *);
extern void cleanup();
extern o_stream *defs_stream, *code_stream;
extern o_stream *nheader_stream, *ncode_stream;
extern void init_types();
extern void multiple_def(const char *);
extern void multiple_def(Symbol *);
extern void undefined(const char *);
extern void undefined(Symbol *);
Symbol *check_var(Symbol *);
Symbol *check_class(Symbol *);
SV_rule *add_rule(Symbol *,SV_rule *);
extern ty_record *sproblem;
extern SV_dflow *create_dflow(Symbol *,ty_record *);
extern int generalized;
extern char *defs_file, *code_file;
Symbol *check_path(Symbol *);
Symbol *check_flow(Symbol *);
%}

/*

A token has an associated value, one of the following types:

1. y_code is a block of C code.

2. y_cstring is a standard null-termindated C string.  Since it is
kept in the heap, its storage should be explicitly freed when there
are no longer any references to it.

3. y_string is a segented string.

4. y_sym is a symbol table entry

*/
%union{
    int y_int;
    C_code *y_code;
    char *y_cstring;
    String *y_string;
    Symbol *y_sym;
    Symbol_list *y_symbol_list;
    ty_record_list *y_record_list;
    SV_rule *y_rule;
    SV_type *y_type;
    ty_record *y_record_type;
    Q_binder *y_q_binder;
}
%token LOCAL RULES END PROLOG REDUCE SIMPLIFIER
%token PROBLEM NODE_BASE NODE_TYPE
%token VALUE_BASE FLOW_VALUE FLOW_BASE FLOW_FUNCTION NEW_VAR 
%token NEW_VALUE MEET
%token COPY_VAR COPY_VALUE FLOW_MAP PROCESS_FUNCTION
%token IN_ACTION OUT_ACTION
%token ID CBLOCK STRING LARROW RARROW AND INIT INSERT ABSORBS VARIABLES CLASS
%token APPLY ITERATIVE PATH JOINS CREATE STAR FLOW_MAP
%token IDENTITY
%type<y_rule> rule rule2 create absorption star join reductions
%type<y_sym> ID var_def q_name node_spec_header
%type<y_symbol_list> var_list
%type<y_code> CBLOCK local_part reduction_code process_part
%type<y_record_type> skel_base problem_base
%type<y_record_list> base_list
%type<y_int> reducer
%type<y_q_binder> q_binder
%start spec
%%

/*##
Main Body of the Grammar

A Sharlit specification describes a data flow problem.  A problem has
a name that will eventually becomes the name of the solver class
generated by Sharlit.  A specification has the following six parts:

- A per-problem section defines information that is common
to all nodes for a particular problem;

- A skeleton section defines information that is stored in
the skeleton structure (see @Snode@ class definition.);

- A variable section defines the data flow variables whose
values are to be found by solving for the data flow problem; and

- A local base or @L@ section defines the super class of all
the nodes defined in the {\em interpreter} section.
These three sections are followed by an {\em interpreter} that defines
all the node types and how they transform the data flow variables.
Finally, the {\em reductions} define how nodes are to be coalesced
to from other nodes.

*/
spec:
	prolog insert_list dflow insert_list
	;
insert_list:
        | insert_list decl
        | insert_list insert
	;
dflow:
	problem_name '='
	per_problem
	flow_map
	node_base
	value_base
	flow_base
	new_var_decl
	copy_var_decl
	meet_decl
	interpreter
	reducer
	END ';'
	{
	    if(n_errors==0){
		problem->generate();
		/* trial */
		if(generalized){
		    problem->code_gen(CGC_header,nheader_stream);
		    problem->code_gen(CGC_code,ncode_stream);
		}
	    }
	    cleanup();
	}
	| error END ';'
	{
		yyerror("bad specification ignored");
	};


problem_name:
	ID problem_base
        {
#define DECL_CLASS(x)							      \
	    defs_stream->put_string("class ");				      \
	    defs_stream->put_string($1->name);				      \
	    defs_stream->put_string(x);					      \
	    defs_stream->put_string(";\n");

	    extern const char *SS_suffix, *LL_suffix, *VV_suffix;

	    if(!$1->loose())
	      multiple_def($1);
	    symtab->enter($1);
	    $1->value=create_dflow($1,$2);
	    set_name($1);

		/*
		  Forward declarations for classes
		  */
	    DECL_CLASS(SS_suffix);
	    DECL_CLASS(LL_suffix);
	    DECL_CLASS(VV_suffix);

		/*
		  Include the generated include file at this point
		  in the code file.
		  */
	    code_stream->put_string("#include \"");
	    code_stream->put_string(defs_file);
	    code_stream->put_string("\"\n");
	    if(generalized){
		ncode_stream->put_string("#include <nsharlit/sharlit.h>\n"
					 "#include \"");
		ncode_stream->put_string(defs_file);
		ncode_stream->put_string("\"\n");
	    }
#undef DECL_CLASS

	};

per_problem:
	PROBLEM ':' CBLOCK ';' {
		assert(problem!=0);
		problem->create_solver_type($3);
	}
	|	/* empty */
	;

flow_map:
	FLOW_MAP ':' CBLOCK ';' {
		problem->flow_map=$3;
	}
	|	/* empty */
	;

/*

This productions parses the definitions of data flow information
that is common to all nodes.  The CBLOCK value is turned
into a C++ class definition.

*/
value_base:
	VALUE_BASE ':'  CBLOCK ';'
	{
		assert(problem!=0);
		problem->define_VV(0,$3);
	}

        | FLOW_VALUE '=' CBLOCK ';'
        {
		assert(problem!=0);
		problem->define_VV(0,$3);
	}

	| VARIABLES '<'ID CBLOCK ';' var_list '>' ';'
	{
		assert(problem!=0);
		problem->define_VV($3,$4);
		problem->extra_classes.append($6);
		delete $6;
	}
	;
var_list:
	{ $$=new Symbol_list(); }
	| var_list var_def {
		$1->enter($2);
	}
	;
var_def:	ID ':' base_list CBLOCK ';'
	{
		if(!$1->loose())
			multiple_def($1);
		symtab->enter($1);
		$1->value=new ty_record($1,*$3,$4);
		delete $3;
	};
base_list:
	ID { $$ = new ty_record_list((ty_record *)check_class($1)->value); }
	| base_list ID { $1->append((ty_record *)check_class($2)->value); }
	;

/*

Base class of all flow functions

*/
flow_base:
	FLOW_BASE ':' CBLOCK ';' {
		assert(problem!=0);
		problem->define_LL($3);
	}
	|	/* empty */
	;

new_var_decl:
	NEW_VAR ':' CBLOCK ';' {
		problem->define_new_var($3);
	}

	| NEW_VALUE ':' CBLOCK ';' {
		problem->define_new_var($3);
	}

	| /* empty */
	;
copy_var_decl:
	COPY_VAR ':' CBLOCK ';' {
		problem->define_copy_var($3);
	}

	| COPY_VALUE ':' CBLOCK ';' {
		problem->define_copy_var($3);
	}

	| /* empty */
	;
meet_decl:
	MEET ':' CBLOCK ';' {
		problem->define_meet($3);
	}
	| /* empty */
	;
interpreter:
	interpreter node_spec
	| node_spec
	;

/*

A prolog is inserted at the beginning of the code file

*/
prolog:
	PROLOG CBLOCK ';' {
		code_stream->put_code($2);
		code_stream->put('\n');
		if(generalized){
		    ncode_stream->put_code($2);
		    ncode_stream->put('\n');
		}
	    }
	|	/* empty */
	;
insert:
	INSERT CBLOCK ';' {
		code_stream->put_code($2);
		code_stream->put('\n');
		if(generalized){
		    ncode_stream->put_code($2);
		    ncode_stream->put('\n');
		}
	    }
	;

/*##
frag(decl,"@decl@ productions")

*/
decl:
	CLASS ID ';' {
		if(!$2->loose())
			multiple_def($2);
		else{
			symtab->enter($2);
			$2->value=new ty_named($2,new ty_record($2,0));
			defs_stream->put_string("class ");
			defs_stream->put_string($2->name);
			defs_stream->put(';');
			defs_stream->nl();
		}
	}

problem_base:
	{ $$=sproblem; }
	| ':' ITERATIVE {
		/* no action for now */
		$$=0;
	}
	| ':' ID {
		if($2->loose()){
			current_input->errmsg("undefined problem type, %s",
				$2->name);
			$$= sproblem;
		}else if(!TAG_IS(SV_type,$2)
			|| !((SV_type *)$2->value)->is(TY_record)){
			current_input->errmsg("invalid problem type, %s",
				$2->name);
			$$=sproblem;
		}else
			$$=(ty_record *)$2->value;
	};

node_base:
	NODE_BASE skel_base CBLOCK ';' {
		assert(problem!=0);
		problem->define_SS($3,$2);
	}

	| NODE_TYPE '=' ID ';' {
		if(!$3->loose())
			multiple_def($3);
		$3->value=new ty_record($3,0);
		symtab->enter($3);
		assert(problem!=0);
		problem->define_SS(0,(ty_record *)$3->value);
	}

        | error ';' {
	    yyerror("bad node type declaration");
	}

	|	/* empty */
	;
skel_base:
	':' ID {
		if(!$2->loose())
			multiple_def($2);
		$2->value=$$=new ty_record($2,0);
		symtab->enter($2);
	}
	| ':' /* empty */ {
		extern ty_record *snode;
		$$=snode; }
	;

/*##
frag(node_spec,"Node specification")

*/
node_spec:
        node_spec_header
		local_part
		CBLOCK	/* action */
		/* extension part to be defined */
	';'
	{
		$1->value=new SV_node($1,$2,$3);
		symtab->enter($1);
		problem->define_node($1);
	}
        | node_spec_header local_part
	    FLOW_FUNCTION CBLOCK /* apply function */
	    process_part
	    ';'
	{
	    $1->value=new SV_node($1,$2,0,$4,$5);
	    symtab->enter($1);
	    problem->define_node($1);
	}
        | node_spec_header IDENTITY
	    process_part
	    ';'
	{
	    $1->value = new SV_node($1,0,0,0,0);
	    symtab->enter($1);
	    if(problem->identity_function)
		yyerror("Multiple identity node functions\n");
	    else
	        problem->define_identity($1);
	}
        | node_spec_header PATH local_part
	    FLOW_FUNCTION CBLOCK /* apply function */
	    process_part
	    ';'
	{
	    $1->value=new SV_path($1,$3,0,$5,$6);
	    symtab->enter($1);
	    problem->define_node($1);
	}

        | error ';' {
	    yyerror("bad data-flow transformer specification");
        };

node_spec_header:
    ID ':'
{
    if(!$1->loose())
      multiple_def($1);
    $$=$1;
}
    ;

process_part:
    PROCESS_FUNCTION CBLOCK
    {
	$$ = $2; 
    }

    | OUT_ACTION CBLOCK
    {
	$$ = $2;
    }

    | { $$=0; }
    ;


/*

Variables private to a flow function and action routines of
a node.

*/

local_part:
	/* empty */ { $$ = 0; }
	| LOCAL CBLOCK { $$=$2; }
	;

/*##
frag(reducer,"Reduction rules")

*/
reducer:
	REDUCE reductions { $$=1; }
        | SIMPLIFIER reductions { $$=1; }
	| { $$=0; }
	;
reductions:
	reductions rule ';'
	| rule ';'
	;
rule:
	{
		/* create a new scope for the rule */
	    Symbol_list *s=new Symbol_list;
	    symtab->push(s);
	}
	rule2 {
		if($2) { $$=$2; $$->symbols=symtab->pop(); }
		else $$=0;
	}
	;
rule2:
        create
	| absorption
	| star
	| join
	| error { yyerror("bad reduction rule"); }
	;

/*##
frag(rule,"A reduction rule")
24~
*/
q_name:
        ID { $$=$1; }
        | { $$=0; }
        ;
q_binder:
        '(' q_name ':' q_name ',' q_name ',' q_name ')' ':'
	{
	    $$=new Q_binder($2,$4,$6,$8);
	    if($$->check()==0)
	      $$=0;
	}
        | '(' ')' ':' { $$=0; }
        ;
reduction_code:
        '=' CBLOCK
	{ $$=$2; }
        | { $$=0; }
        ;
create:
	q_binder ID CREATE '=' CBLOCK
	{
	    if (check_path($2))
	      $$=add_rule($2,new SVR_create($1,$2,$5));
	    else
	      $$=0;
	}
	| q_binder ID CREATE q_binder ID '=' CBLOCK
	{
	    if (check_path($2) && check_flow($5))
		$$=add_rule($2,new SVR_unit($1,$2,$4,$5,$7));
	    else
	      $$=0;
	}
	;
absorption:
	q_binder ID ABSORBS q_binder ID reduction_code
	{
	    if (check_path($2) && check_flow($5))
		$$=add_rule($2,new SVR_absorb($1,$2,$4,$5,$6));
	    else
	      $$=0;
	}
	;
star:
	q_binder ID STAR '=' CBLOCK
	{
	    if(check_path($2))
	      $$=add_rule($2,new SVR_star($1,$2,$5));
	    else
	      $$=0;
	};
join:
        q_binder ID JOINS q_binder ID '=' CBLOCK
	{
	    if (check_path($2) && check_flow($5))
	      $$=add_rule($2,new SVR_meet($1,$2,$4,$5,$7));
	    else
	      $$=0;
	}
	;

/* end of YACC part */
%%

/*

Flags and stuff

*/
char *input_file;
char *input_base;
char *defs_file = "solver.h";
char *code_file = "solver.cc";
char *define_name;
o_stream *defs_stream, *code_stream;
int line_control;
int test_flag;
int debug_flag;
int generalized;
extern void finish_files(), init_files(), init_code_file(int);

/*##
frag(init_symbols,"Initialize the symbol table")

*/
struct k_init {
	char *name;
	int value;
} keywds[] = {
	{"local", LOCAL}, {"rules",RULES},
	{"end",END},{"prolog",PROLOG},{"reduce",REDUCE},
	{"simplifier",SIMPLIFIER},
	{"prologue",PROLOG},{"and",AND},{"init",INIT},{"insert", INSERT},
	{"variables",VARIABLES},{"class",CLASS},
	{"problem", PROBLEM},{"node_base",NODE_BASE},{"value_base",VALUE_BASE},
	{"flow_base",FLOW_BASE},{"flow_function",FLOW_FUNCTION},
	{"iterative",ITERATIVE},{"new_var",NEW_VAR},{"copy_var",COPY_VAR},
	{"meet",MEET},{"flow_map",FLOW_MAP},
	{"process_function",PROCESS_FUNCTION},
	{"in_action",IN_ACTION},
	{"out_action",OUT_ACTION},
	{"path", PATH},	{"create", CREATE}, {"absorbs", ABSORBS},
	{"joins",JOINS},{"star",STAR}, {"flow_map",FLOW_MAP},
	{"identity",IDENTITY},
	{"flow_value",FLOW_VALUE},{"new_value",NEW_VALUE},
	{"copy_value",COPY_VALUE},
	{"node_type",NODE_TYPE},
	{0,0}
};
void
init_symbols()
{
	k_init *kp;
	symtab=new Symbol_table();
	symtab->push(new Hashed_scope());
	for(kp=keywds;kp->name;kp++){
		Symbol *sp=new Symbol(kp->name);
		symtab->enter(sp);
		sp->value=new SV_keyword(kp->value);
	}
}


/*##
frag(yylex,"The lexical routine called by {\tt yyparse}")

*/
int
yylex()
{
	int i=current_input->yylex();
	return i;
}
void
yyerror(const char *msg)
{
	current_input->errmsg("%s",msg);
}

/*##
frag(utility,"Utility routines")

*/
void
_assert(char *e, char*f,int l)
{
	fprintf(stderr,"assert(%s) failed on line %d of %s\nAborting!\n",
		e,l,f);
	fflush(stderr);
	abort();
}
const char *
const_string(const char *s)
{
	char*name=new char[strlen(s)+1];
	strcpy(name,s);
	return name;
}

/*##
frag(checks,"@check_var@")

*/
Symbol *
check_var(Symbol *sp)
{
	if(symtab->top()->has(sp)){
	    multiple_def(sp);
	    return 0;
	}else if(!sp->loose()){
	    current_input->warning("overriding former definition of %s in scope",
				   sp->name);
	    symtab->enter(sp=new Symbol(sp));
	}else
	  symtab->enter(sp);
	sp->value=0;
	return sp;
}
int
Q_binder::check()
{
    return (Q_name==0 || check_var(Q_name))
      & (V_name==0 || check_var(V_name))
	& (L_name==0 || check_var(L_name))
	  & (S_name==0 || check_var(S_name));
}

/*##
frag(check_class,"@check_class@")

*/
Symbol *
check_class(Symbol *s)
{
	if(s->loose() || !TAG_IS(SV_type,s) || s->value->is(TY_record)){
		current_input->errmsg("%s is not a class\n",s->name);
		return &error_sym;
	}
	return s;
}

/*##
frag(check_node,"@check_node@")

*/

Symbol *
check_node(Symbol *s)
{
    if(s->loose()){
	current_input->errmsg("undefined node type: %s",s->name);
	return 0;
    }else if(!TAG_IS(SV_node,s)){
	current_input->errmsg("%s is not a node type",s->name);
	return 0;
    }else
      return s;
}
Symbol *
check_path(Symbol *s)
{
    if(s->loose()){
	current_input->errmsg("undefined path type: %s",s->name);
	return 0;
    }else if(!s->is(TAG(SV_path))){
	current_input->errmsg("%s is not a path",s->name);
	return 0;
    }else
      return s;
}
Symbol *
check_flow(Symbol *s)
{
    if(s->loose()){
	current_input->errmsg("undefined path type: %s",s->name);
	return 0;
    }else if(!s->is(TAG(SV_path)) && !s->is(TAG(SV_node))){
	current_input->errmsg("%s is not a path nor a node",s->name);
	return 0;
    }else
      return s;
}


/*##
frag(multiple_def,"Multiple Definitions error message")

*/
void
multiple_def(Symbol *s)
{
	multiple_def(s->name);
}
void
multiple_def(const char *n)
{
	current_input->errmsg("multiple definitions of %s",n);
}
void
undefined(Symbol *s)
{
	undefined(s->name);
}
void
undefined(const char *n)
{
	current_input->errmsg("%s is undefined",n);
}

/*##
frag(add_rule,"@add_rule@")

*/
SV_rule *
add_rule(Symbol *s,SV_rule *new_rule)
{
    SV_rule *r;
    r=TAG_CHECK(SV_path,s)->add_rule(new_rule);
    if(r){
	problem->rules.append(r);
	return r;
    }
    current_input->errmsg("duplicate rule");
    delete new_rule;
    return 0;
}

int main(int, char **argv)
{
    cmdnam=argv[0];
    for(argv++;*argv;argv++)
      if(argv[0][0]=='-' && argv[0][1]!=0)
	switch(argv[0][1]){
	  case 'd':
	    debug_flag=1;
	    break;
	  case 'L':
	    line_control=1;
	    break;
	  case 'T':
	    test_flag=1;
	    break;
	  case 'g':
	    generalized=1;
	    break;
	  default:
	    fprintf(stderr,"Unknown switch: %s\n",argv[0]);
	    exit(1);
	}
      else if(input_file==0)
	input_file=argv[0];
      else {
	  fprintf(stderr,"%s:only one file at a time, please\n",
		  cmdnam);
	  exit(1);
      }
    if(input_file==0){
	fprintf(stderr,"%s: no input file given\n",cmdnam);
	exit(1);
    }
    init_symbols();
    init_types();
    if(strcmp(input_file,"-")==0){
	current_input = new f_i_stream(stdin,"-");
	defs_file=code_file="-";
	defs_stream=new f_o_stream(stdout,"-");
	code_stream=new f_o_stream(stdout,"-");
	input_base="_STDIN_";
    }else{
	char *tmp, *junk;
	tmp=strrchr(input_file,'.');
	if(tmp==0 || (strcmp(tmp,".dflow")!=0 && strcmp(tmp,".odflow")!=0)){
	    fprintf(stderr,"%s: input must have .dflow ending\n",
		    cmdnam);
	    exit(1);
	}

	    /*
	      Copy input_file into junk, strip away all but the
	      basename then set input_base to the basename.
	      */
	junk = new char[strlen(input_file)+1];
	strcpy(junk,input_file);
	tmp = strrchr(junk,'.'); // tmp!=0 because of earlier tests
	*tmp = 0;
	input_base = strrchr(junk,'/');
	if(input_base==NULL)           // zero if no slash.
	  input_base=junk;
	
	current_input=new f_i_stream(input_file);
	defs_stream=new f_o_stream(defs_file);
	code_stream=new f_o_stream(code_file);
	if(generalized){
	    nheader_stream=new f_o_stream("debug.h");
	    ncode_stream=new f_o_stream("debug.c");
	}
    }
    init_files();
    init_code_file(test_flag);
    yyparse();
    finish_files();
    delete code_stream;
    delete defs_stream;
    if(generalized){
	delete nheader_stream;
	delete ncode_stream;
    }
    exit(n_errors>0);
    return 0;
}
